# CPUID feature bits, from LSB to MSB:
# (Names and descriptions gathered from various Intel and AMD sources)

cap_raw = (
# EAX=01H ECX:
"""SSE3
PCLMULQDQ
DTES64: 64-bit debug store
MONITOR: MONITOR/MWAIT
DS-CPL: CPL qualified debug store
VMX: virtual machine extensions
SMX: safer mode extensions
EST: enhanced SpeedStep
TM2: thermal monitor 2
SSSE3
CNXT-ID: L1 context ID
?(ecx11)
FMA: fused multiply add
CMPXCHG16B
xTPR: xTPR update control
PDCM: perfmon and debug capability
?(ecx16)
PCID: process context identifiers
DCA: direct cache access
SSE4_1
SSE4_2
x2APIC: extended xAPIC support
MOVBE
POPCNT
TSC-DEADLINE
AES
XSAVE: XSAVE instructions supported
OSXSAVE: XSAVE instructions enabled
AVX
F16C: half-precision convert
?(ecx30)
RAZ: used by hypervisor to indicate guest status
""" +

# EAX=01H EDX:
"""FPU
VME: virtual 8086 mode enhancements
DE: debugging extension
PSE: page size extension
TSC: time stamp counter
MSR: model specific registers
PAE: physical address extension
MCE: machine-check exception
CMPXCHG8
APIC
?(edx10)
SEP: fast system call
MTRR: memory type range registers
PGE: page global enable
MCA: machine-check architecture
CMOV
PAT: page attribute table
PSE-36: 36-bit page size extension
PSN: processor serial number
CLFSH: CLFLUSH
?(edx20)
DS: debug store
ACPI
MMX
FXSR: FXSAVE and FXSTOR
SSE
SSE2
SS: self-snoop
HTT: hyper-threading
TM: thermal monitor
?(edx30)
PBE: pending break enable
""" +

# EAX=80000001H ECX:
"""LAHF: LAHF/SAHF instructions
CMP: core multi-processing legacy mode
SVM: secure virtual machine
ExtApic
AltMovCr8
ABM: LZCNT instruction
SSE4A
MisAlignSse
3DNowPrefetch
OSVW: OS visible workaround
IBS: instruction based sampling
XOP: extended operation support
SKINIT
WDT: watchdog timer support
?(ext:ecx14)
LWP: lightweight profiling support
FMA4: 4-operand FMA
?(ext:ecx17)
?(ext:ecx18)
NodeId
?(ext:ecx20)
TBM: trailing bit manipulation extensions
TopologyExtensions
?(ext:ecx23)
?(ext:ecx24)
?(ext:ecx25)
?(ext:ecx26)
?(ext:ecx27)
?(ext:ecx28)
?(ext:ecx29)
?(ext:ecx30)
?(ext:ecx31)
""" +

# EAX=80000001H ECX:
"""FPU[2]
VME[2]
DE[2]
PSE[2]
TSC[2]
MSR[2]
PAE[2]
MCE[2]
CMPXCHG8[2]
APIC[2]
?(ext:edx10)
SYSCALL: SYSCALL/SYSRET instructions
MTRR[2]
PGE[2]
MCA[2]
CMOV[2]
PAT[2]
PSE36[2]
?(ext:edx18)
MP: MP-capable
NX: no execute bit
?(ext:edx21)
MmxExt
MMX[2]
FXSR[2]
FFXSR
1GB: 1GB pages
RDTSCP
?(ext:edx28)
x86-64
3DNowExt
3DNow
"""
)

cap_bits = []
cap_descs = {}
idx = 0
for c in cap_raw.strip().split('\n'):
    s = c.split(':')
    if len(s) == 1:
        cap_bits.append((s[0], None, idx))
    else:
        cap_bits.append((s[0], s[1], idx))
        cap_descs[s[0]] = s[1]
    idx += 1



from django.conf.urls.defaults import *

from django.contrib import admin
admin.autodiscover()

urlpatterns = patterns('',
    (r'^$', 'userreport.views.index'),
    (r'^report/', include('userreport.urls')),
    (r'^private/', include('userreport.urls_private')),
    (r'^admin/', include(admin.site.urls)),
)



from django.db import models
from django.utils import simplejson

import re

class UserReport(models.Model):

    uploader = models.IPAddressField(editable = False)

    # Hex SHA-1 digest of user's reported ID
    # (The hashing means that publishing the database won't let people upload
    # faked reports under someone else's user ID, and also ensures a simple
    # consistent structure)
    user_id_hash = models.CharField(max_length = 40, db_index = True, editable = False)

    # When the server received the upload
    upload_date = models.DateTimeField(auto_now_add = True, db_index = True, editable = False)

    # When the user claims to have generated the report
    generation_date = models.DateTimeField(editable = False)

    data_type = models.CharField(max_length = 16, db_index = True, editable = False)

    data_version = models.IntegerField(editable = False)

    data = models.TextField(editable = False)

    def data_json(self):
        if not hasattr(self, 'cached_json'):
            try:
                self.cached_json = simplejson.loads(self.data)
            except:
                self.cached_json = None
        return self.cached_json

    def data_json_nocache(self):
        try:
            return simplejson.loads(self.data)
        except:
            return None

    def clear_cache(self):
        delattr(self, 'cached_json')

    def downcast(self):
        if self.data_type == 'hwdetect':
            return UserReport_hwdetect.objects.get(id = self.id)
        else:
            return self

class UserReport_hwdetect(UserReport):

    class Meta:
        proxy = True

    def os(self):
        os = 'Unknown'
        json = self.data_json()
        if json:
            if json['os_win']:
                os = 'Windows'
            elif json['os_linux']:
                os = 'Linux'
            elif json['os_macosx']:
                os = 'OS X'
        return os

    def gl_renderer(self):
        json = self.data_json()
        if json is None or 'GL_RENDERER' not in json:
            return None
        # The renderer string should typically be interpreted as UTF-8
        try:
            return json['GL_RENDERER'].encode('iso-8859-1').decode('utf-8').strip()
        except UnicodeError:
            return json['GL_RENDERER'].strip()

    def gl_extensions(self):
        json = self.data_json()
        if json is None or 'GL_EXTENSIONS' not in json:
            return None
        vals = re.split(r'\s+', json['GL_EXTENSIONS'])
        return frozenset(v for v in vals if len(v)) # skip empty strings (e.g. no extensions at all, or leading/trailing space)

    def gl_limits(self):
        json = self.data_json()
        if json is None:
            return None

        limits = {}
        for (k, v) in json.items():
            if not k.startswith('GL_'):
                continue
            if k == 'GL_VERSION':
                m = re.match(r'^(\d+\.\d+).*', v)
                if m:
                    limits[k] = '%s [...]' % m.group(1)
                continue
            if k in ('GL_RENDERER', 'GL_EXTENSIONS'):
                continue
            # Hide some values that got deleted from the report in r8953, for consistency
            if k in ('GL_MAX_COLOR_MATRIX_STACK_DEPTH', 'GL_FRAGMENT_PROGRAM_ARB.GL_MAX_PROGRAM_ADDRESS_REGISTERS_ARB', 'GL_FRAGMENT_PROGRAM_ARB.GL_MAX_PROGRAM_NATIVE_ADDRESS_REGISTERS_ARB'):
                continue
            # Hide some pixel depth values that are not really correlated with device
            if k in ('GL_RED_BITS', 'GL_GREEN_BITS', 'GL_BLUE_BITS', 'GL_ALPHA_BITS', 'GL_INDEX_BITS', 'GL_DEPTH_BITS', 'GL_STENCIL_BITS',
                     'GL_ACCUM_RED_BITS', 'GL_ACCUM_GREEN_BITS', 'GL_ACCUM_BLUE_BITS', 'GL_ACCUM_ALPHA_BITS'):
                continue
            limits[k] = v
        return limits

    # Construct a nice-looking concise graphics device identifier
    # (skipping boring hardware/driver details)
    def gl_device_identifier(self):
        r = self.gl_renderer()
        m = re.match(r'^(?:AMD |ATI |NVIDIA |Mesa DRI )?(.*?)\s*(?:GEM 20100328 2010Q1|GEM 20100330 DEVELOPMENT|GEM 20091221 2009Q4|20090101|Series)?\s*(?:x86|/AGP|/PCI|/MMX|/MMX\+|/SSE|/SSE2|/3DNOW!|/3DNow!|/3DNow!\+)*(?: TCL| NO-TCL)?(?: DRI2)?(?: \(Microsoft Corporation - WDDM\))?(?: OpenGL Engine)?\s*$', r)
        if m:
            r = m.group(1)
        return r.strip()

    def gl_vendor(self):
        json = self.data_json()
        return json['GL_VENDOR'].strip()

    # Construct a nice string identifying the driver
    def gl_driver(self):
        json = self.data_json()
        gfx_drv_ver = json['gfx_drv_ver']

        # Try the Mesa git style first
        m = re.match(r'^OpenGL \d+\.\d+(?:\.\d+)? (Mesa \d+\.\d+)-devel \(git-([a-f0-9]+)', gfx_drv_ver)
        if m:
            return '%s-git-%s' % (m.group(1), m.group(2))

        # Try the normal Mesa style
        m = re.match(r'^OpenGL \d+\.\d+(?:\.\d+)? (Mesa .*)$', gfx_drv_ver)
        if m:
            return m.group(1)

        # Try the NVIDIA Linux style
        m = re.match(r'^OpenGL \d+\.\d+(?:\.\d+)? NVIDIA (.*)$', gfx_drv_ver)
        if m:
            return m.group(1)

        # Try the ATI Catalyst Linux style
        m = re.match(r'^OpenGL (\d+\.\d+\.\d+) Compatibility Profile Context(?: FireGL)?$', gfx_drv_ver)
        if m:
            return m.group(1)

        # Try the non-direct-rendering ATI Catalyst Linux style
        m = re.match(r'^OpenGL 1\.4 \((\d+\.\d+\.\d+) Compatibility Profile Context(?: FireGL)?\)$', gfx_drv_ver)
        if m:
            return '%s (indirect)' % m.group(1)

        # Try to guess the relevant Windows driver
        # (These are the ones listed in lib/sysdep/os/win/wgfx.cpp)

        if json['GL_VENDOR'] == 'NVIDIA Corporation':
            # Assume 64-bit takes precedence
            m = re.search(r'nvoglv64.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)
            m = re.search(r'nvoglv32.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)
            m = re.search(r'nvoglnt.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)

        if json['GL_VENDOR'] in ('ATI Technologies Inc.', 'Advanced Micro Devices, Inc.'):
            m = re.search(r'atioglxx.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)
            m = re.search(r'atioglx2.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)
            m = re.search(r'atioglaa.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)

        if json['GL_VENDOR'] == 'Intel':
            # Assume 64-bit takes precedence
            m = re.search(r'ig4icd64.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)
            m = re.search(r'ig4icd32.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)
            # Legacy 32-bit
            m = re.search(r'iglicd32.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)
            m = re.search(r'ialmgicd32.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)
            m = re.search(r'ialmgicd.dll \((.*?)\)', gfx_drv_ver)
            if m: return m.group(1)

        return gfx_drv_ver


class GraphicsDevice(models.Model):
    device_name = models.CharField(max_length = 128, db_index = True)
    vendor = models.CharField(max_length = 64)
    renderer = models.CharField(max_length = 128)
    os = models.CharField(max_length = 16)
    driver = models.CharField(max_length = 128)
    usercount = models.IntegerField()

class GraphicsExtension(models.Model):
    device = models.ForeignKey(GraphicsDevice)
    name = models.CharField(max_length = 128, db_index = True)

class GraphicsLimit(models.Model):
    device = models.ForeignKey(GraphicsDevice)
    name = models.CharField(max_length = 128, db_index = True)
    value = models.CharField(max_length = 64)




# http://code.djangoproject.com/attachment/ticket/5908/cycle.py

from django.utils.translation import ungettext, ugettext as _
from django.utils.encoding import force_unicode
from django import template
from django.template import defaultfilters
from django.template import Node, Variable
from django.conf import settings
from itertools import cycle as itertools_cycle

register = template.Library()


   
class SafeCycleNode(Node):
    def __init__(self, cyclevars, variable_name=None):
        self.cyclevars = cyclevars
        self.cycle_iter = itertools_cycle(cyclevars)
        self.variable_name = variable_name

    def render(self, context):
        if context.has_key('forloop'):
            if not context.get(self):
                context[self] = True
                self.cycle_iter = itertools_cycle(self.cyclevars)
        value = self.cycle_iter.next()
        value = Variable(value).resolve(context)
        if self.variable_name:
            context[self.variable_name] = value
        return value



#@register.tag
def safe_cycle(parser, token):
    """
    Cycles among the given strings each time this tag is encountered.

    Within a loop, cycles among the given strings each time through
    the loop::

        {% for o in some_list %}
            <tr class="{% cycle 'row1' 'row2' %}">
                ...
            </tr>
        {% endfor %}

    Outside of a loop, give the values a unique name the first time you call
    it, then use that name each sucessive time through::

            <tr class="{% cycle 'row1' 'row2' 'row3' as rowcolors %}">...</tr>
            <tr class="{% cycle rowcolors %}">...</tr>
            <tr class="{% cycle rowcolors %}">...</tr>

    You can use any number of values, seperated by spaces. Commas can also
    be used to separate values; if a comma is used, the cycle values are
    interpreted as literal strings.
    """

    # Note: This returns the exact same node on each {% cycle name %} call;
    # that is, the node object returned from {% cycle a b c as name %} and the
    # one returned from {% cycle name %} are the exact same object.  This
    # shouldn't cause problems (heh), but if it does, now you know.
    #
    # Ugly hack warning: this stuffs the named template dict into parser so
    # that names are only unique within each template (as opposed to using
    # a global variable, which would make cycle names have to be unique across
    # *all* templates.

    args = token.split_contents()

    if len(args) < 2:
        raise TemplateSyntaxError("'cycle' tag requires at least two arguments")

    if ',' in args[1]:
        # Backwards compatibility: {% cycle a,b %} or {% cycle a,b as foo %}
        # case.
        args[1:2] = ['"%s"' % arg for arg in args[1].split(",")]

    if len(args) == 2:
        # {% cycle foo %} case.
        name = args[1]
        if not hasattr(parser, '_namedCycleNodes'):
            raise TemplateSyntaxError("No named cycles in template."
                                      " '%s' is not defined" % name)
        if not name in parser._namedCycleNodes:
            raise TemplateSyntaxError("Named cycle '%s' does not exist" % name)
        return parser._namedCycleNodes[name]

    if len(args) > 4 and args[-2] == 'as':
        name = args[-1]
        node = SafeCycleNode(args[1:-2], name)
        if not hasattr(parser, '_namedCycleNodes'):
            parser._namedCycleNodes = {}
        parser._namedCycleNodes[name] = node
    else:
        node = SafeCycleNode(args[1:])
    return node
safe_cycle = register.tag(safe_cycle)



from userreport.models import UserReport, GraphicsDevice, GraphicsExtension, GraphicsLimit
from django.contrib import admin

class UserReportAdmin(admin.ModelAdmin):

    readonly_fields = ['uploader', 'user_id_hash', 'upload_date', 'generation_date', 'data_type', 'data_version', 'data']
    fieldsets = [
        ('User', {'fields': ['uploader', 'user_id_hash']}),
        ('Dates', {'fields': ['upload_date', 'generation_date']}),
        (None, {'fields': ['data_type', 'data_version', 'data']}),
    ]
    list_display = ('uploader', 'user_id_hash', 'data_type', 'data_version', 'upload_date', 'generation_date')
    list_filter = ['upload_date', 'generation_date', 'data_type']
    search_fields = ['=uploader', '=user_id_hash', 'data']
    date_hierarchy = 'upload_date'

class GraphicsDeviceAdmin(admin.ModelAdmin):
    pass

class GraphicsExtensionAdmin(admin.ModelAdmin):
    pass

class GraphicsLimitAdmin(admin.ModelAdmin):
    pass

admin.site.register(UserReport, UserReportAdmin)
admin.site.register(GraphicsDevice, GraphicsDeviceAdmin)
admin.site.register(GraphicsExtension, GraphicsExtensionAdmin)
admin.site.register(GraphicsLimit, GraphicsLimitAdmin)




# Adapted from:
#   http://enichan.darksiren.net/wordpress/?p=49
# which was adapted from some version of
#   https://devel.nuclex.org/framework/browser/game/Nuclex.Game/trunk/Source/Packing/CygonRectanglePacker.cs
# which has the following license:
#
#   Copyright (C) 2002-2009 Nuclex Development Labs
#
#   This library is free software; you can redistribute it and/or
#   modify it under the terms of the IBM Common Public License as
#   published by the IBM Corporation; either version 1.0 of the
#   License, or (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   IBM Common Public License for more details.

from bisect import bisect_left

class OutOfSpaceError(Exception): pass

class Point(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y
 
    def __cmp__(self, other):
        """Compares the starting position of height slices"""
        return self.x - other.x

class RectanglePacker(object):
    """Base class for rectangle packing algorithms
 
    By uniting all rectangle packers under this common base class, you can
    easily switch between different algorithms to find the most efficient or
    performant one for a given job.
 
    An almost exhaustive list of packing algorithms can be found here:
    http://www.csc.liv.ac.uk/~epa/surveyhtml.html"""
 
    def __init__(self, packingAreaWidth, packingAreaHeight):
        """Initializes a new rectangle packer
 
        packingAreaWidth: Maximum width of the packing area
        packingAreaHeight: Maximum height of the packing area"""
        self.packingAreaWidth = packingAreaWidth
        self.packingAreaHeight = packingAreaHeight
 
    def Pack(self, rectangleWidth, rectangleHeight):
        """Allocates space for a rectangle in the packing area
 
        rectangleWidth: Width of the rectangle to allocate
        rectangleHeight: Height of the rectangle to allocate
 
        Returns the location at which the rectangle has been placed"""
        point = self.TryPack(rectangleWidth, rectangleHeight)
 
        if not point:
            raise OutOfSpaceError("Rectangle does not fit in packing area")
 
        return point
 
    def TryPack(self, rectangleWidth, rectangleHeight):
        """Tries to allocate space for a rectangle in the packing area
 
        rectangleWidth: Width of the rectangle to allocate
        rectangleHeight: Height of the rectangle to allocate
 
        Returns a Point instance if space for the rectangle could be allocated
        be found, otherwise returns None"""
        raise NotImplementedError

class DumbRectanglePacker(RectanglePacker):
    def __init__(self, packingAreaWidth, packingAreaHeight):
        RectanglePacker.__init__(self, packingAreaWidth, packingAreaHeight)
        self.x = 0
        self.y = 0
        self.rowh = 0

    def TryPack(self, rectangleWidth, rectangleHeight):
        if self.x + rectangleWidth >= self.packingAreaWidth:
            self.x = 0
            self.y += self.rowh
            self.rowh = 0
            if self.y + rectangleHeight >= self.packingAreaHeight:
                return None

        r = Point(self.x, self.y)
        self.x += rectangleWidth
        self.rowh = max(self.rowh, rectangleHeight)
        return r

class CygonRectanglePacker(RectanglePacker):
    """
    Packer using a custom algorithm by Markus 'Cygon' Ewald
 
    Algorithm conceived by Markus Ewald (cygon at nuclex dot org), though
    I'm quite sure I'm not the first one to come up with it :)
 
    The algorithm always places rectangles as low as possible in the packing
    area. So, for any new rectangle that is to be added, the packer has to
    determine the X coordinate at which the rectangle can have the lowest
    overall height without intersecting any other rectangles.
 
    To quickly discover these locations, the packer uses a sophisticated
    data structure that stores the upper silhouette of the packing area. When
    a new rectangle needs to be added, only the silouette edges need to be
    analyzed to find the position where the rectangle would achieve the lowest"""
 
    def __init__(self, packingAreaWidth, packingAreaHeight):
        """Initializes a new rectangle packer
 
        packingAreaWidth: Maximum width of the packing area
        packingAreaHeight: Maximum height of the packing area"""
        RectanglePacker.__init__(self, packingAreaWidth, packingAreaHeight)
 
        # Stores the height silhouette of the rectangles
        self.heightSlices = []
 
        # At the beginning, the packing area is a single slice of height 0
        self.heightSlices.append(Point(0,0))
 
    def TryPack(self, rectangleWidth, rectangleHeight):
        """Tries to allocate space for a rectangle in the packing area
 
        rectangleWidth: Width of the rectangle to allocate
        rectangleHeight: Height of the rectangle to allocate
 
        Returns a Point instance if space for the rectangle could be allocated
        be found, otherwise returns None"""
        placement = None
 
        # If the rectangle is larger than the packing area in any dimension,
        # it will never fit!
        if rectangleWidth > self.packingAreaWidth or rectangleHeight > \
        self.packingAreaHeight:
            return None
 
        # Determine the placement for the new rectangle
        placement = self.tryFindBestPlacement(rectangleWidth, rectangleHeight)
 
        # If a place for the rectangle could be found, update the height slice
        # table to mark the region of the rectangle as being taken.
        if placement:
            self.integrateRectangle(placement.x, rectangleWidth, placement.y \
            + rectangleHeight)
 
        return placement
 
    def tryFindBestPlacement(self, rectangleWidth, rectangleHeight):
        """Finds the best position for a rectangle of the given dimensions
 
        rectangleWidth: Width of the rectangle to find a position for
        rectangleHeight: Height of the rectangle to find a position for
 
        Returns a Point instance if a valid placement for the rectangle could
        be found, otherwise returns None"""
        # Slice index, vertical position and score of the best placement we
        # could find
        bestSliceIndex = -1 # Slice index where the best placement was found
        bestSliceY = 0 # Y position of the best placement found
        # lower == better!
        bestScore = self.packingAreaHeight 

        # This is the counter for the currently checked position. The search
        # works by skipping from slice to slice, determining the suitability
        # of the location for the placement of the rectangle.
        leftSliceIndex = 0
 
        # Determine the slice in which the right end of the rectangle is located
        rightSliceIndex = bisect_left(self.heightSlices, Point(rectangleWidth, 0))
 
        while rightSliceIndex <= len(self.heightSlices):
            # Determine the highest slice within the slices covered by the
            # rectangle at its current placement. We cannot put the rectangle
            # any lower than this without overlapping the other rectangles.
            highest = self.heightSlices[leftSliceIndex].y
            for index in xrange(leftSliceIndex + 1, rightSliceIndex):
                if self.heightSlices[index].y > highest:
                    highest = self.heightSlices[index].y
 
            # Only process this position if it doesn't leave the packing area
            if highest + rectangleHeight < self.packingAreaHeight:
                score = highest
 
                if score < bestScore:
                    bestSliceIndex = leftSliceIndex
                    bestSliceY = highest
                    bestScore = score
 
            # Advance the starting slice to the next slice start
            leftSliceIndex += 1
            if leftSliceIndex >= len(self.heightSlices):
                break
 
            # Advance the ending slice until we're on the proper slice again,
            # given the new starting position of the rectangle.
            rightRectangleEnd = self.heightSlices[leftSliceIndex].x + rectangleWidth
            while rightSliceIndex <= len(self.heightSlices):
                if rightSliceIndex == len(self.heightSlices):
                    rightSliceStart = self.packingAreaWidth
                else:
                    rightSliceStart = self.heightSlices[rightSliceIndex].x
 
                # Is this the slice we're looking for?
                if rightSliceStart > rightRectangleEnd:
                    break
 
                rightSliceIndex += 1
 
            # If we crossed the end of the slice array, the rectangle's right
            # end has left the packing area, and thus, our search ends.
            if rightSliceIndex > len(self.heightSlices):
                break
 
        # Return the best placement we found for this rectangle. If the
        # rectangle didn't fit anywhere, the slice index will still have its
        # initialization value of -1 and we can report that no placement
        # could be found.
        if bestSliceIndex == -1:
            return None
        else:
            return Point(self.heightSlices[bestSliceIndex].x, bestSliceY)
 
    def integrateRectangle(self, left, width, bottom):
        """Integrates a new rectangle into the height slice table
 
        left: Position of the rectangle's left side
        width: Width of the rectangle
        bottom: Position of the rectangle's lower side"""
        # Find the first slice that is touched by the rectangle
        startSlice = bisect_left(self.heightSlices, Point(left, 0))
 
        # We scored a direct hit, so we can replace the slice we have hit
        firstSliceOriginalHeight = self.heightSlices[startSlice].y
        self.heightSlices[startSlice] = Point(left, bottom)
 
        right = left + width
        startSlice += 1
 
        # Special case, the rectangle started on the last slice, so we cannot
        # use the start slice + 1 for the binary search and the possibly
        # already modified start slice height now only remains in our temporary
        # firstSliceOriginalHeight variable
        if startSlice >= len(self.heightSlices):
            # If the slice ends within the last slice (usual case, unless it
            # has the exact same width the packing area has), add another slice
            # to return to the original height at the end of the rectangle.
            if right < self.packingAreaWidth:
                self.heightSlices.append(Point(right, firstSliceOriginalHeight))
        else: # The rectangle doesn't start on the last slice
            endSlice = bisect_left(self.heightSlices, Point(right,0), \
            startSlice, len(self.heightSlices))
 
            # Another direct hit on the final slice's end?
            if endSlice < len(self.heightSlices) and not (Point(right, 0) < self.heightSlices[endSlice]):
                del self.heightSlices[startSlice:endSlice]
            else: # No direct hit, rectangle ends inside another slice
                # Find out to which height we need to return at the right end of
                # the rectangle
                if endSlice == startSlice:
                    returnHeight = firstSliceOriginalHeight
                else:
                    returnHeight = self.heightSlices[endSlice - 1].y
 
                # Remove all slices covered by the rectangle and begin a new
                # slice at its end to return back to the height of the slice on
                # which the rectangle ends.
                del self.heightSlices[startSlice:endSlice]
                if right < self.packingAreaWidth:
                    self.heightSlices.insert(startSlice, Point(right, returnHeight))



from userreport.models import UserReport, UserReport_hwdetect

from django.http import HttpResponse, HttpResponseForbidden
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.core.paginator import Paginator, InvalidPage, EmptyPage
from django.db.models import Q
from django.utils import simplejson

import re
import numpy

class hashabledict(dict):
    def __hash__(self):
        return hash(tuple(sorted(self.items())))

def render_reports(request, reports, template, args):
    paginator = Paginator(reports, args.get('pagesize', 100))
    try:
        page = int(request.GET.get('page', '1'))
    except ValueError:
        page = 1
    try:
        report_page = paginator.page(page)
    except (EmptyPage, InvalidPage):
        report_page = paginator.page(paginator.num_pages)

    args['report_page'] = report_page
    return render_to_response(template, args)


def report_user(request, user):
    reports = UserReport.objects.order_by('-upload_date')
    reports = reports.filter(user_id_hash = user)

    return render_reports(request, reports, 'reports/user.html', {'user': user})

def report_messages(request):
    reports = UserReport.objects.order_by('-upload_date')
    reports = reports.filter(data_type = 'message', data_version__gte = 1)

    return render_reports(request, reports, 'reports/message.html', {})

def report_profile(request):
    reports = UserReport.objects.order_by('-upload_date')
    reports = reports.filter(data_type = 'profile', data_version__gte = 2)

    return render_reports(request, reports, 'reports/profile.html', {'pagesize': 20})

def report_hwdetect(request):
    reports = UserReport.objects.order_by('-upload_date')
    reports = reports.filter(data_type = 'hwdetect', data_version__gte = 1)

    return render_reports(request, reports, 'reports/hwdetect.html', {})

def report_gfx(request):
    reports = UserReport.objects.order_by('-upload_date')
    reports = reports.filter(data_type = 'hwdetect', data_version__gte = 1)

    return render_reports(request, reports, 'reports/gfx.html', {'pagesize': 1000})

def report_performance(request):
    reports = UserReport.objects.order_by('upload_date')
    reports = reports.filter(
        Q(data_type = 'hwdetect', data_version__gte = 5) |
        Q(data_type = 'profile', data_version__gte = 2)
    )
    #reports = reports[:500]

    def summarise_hwdetect(report):
        json = report.data_json_nocache()
        return {
            'cpu_identifier': json['cpu_identifier'],
            'device': report.gl_device_identifier(),
            'build_debug': json['build_debug'],
            'build_revision': json['build_revision'],
            'build_datetime': json['build_datetime'],
            'gfx_res': (json['video_xres'], json['video_yres']),
        }

    def summarise_profile(report):
        json = report.data_json_nocache()
        if 'map' in json:
            mapname = json['map']
        else:
            mapname = 'unknown' # e.g. random maps
        msecs = None
        shadows = False
        fancywater = False
        for name,table in json['profiler'].items():
            m = re.match(r'Profiling Information for: root \(Time in node: (\d+\.\d+) msec/frame\)', name)
            if m:
                try:
                    msecs = float(table['data']['render'][2])
                except KeyError:
                    pass

                try:
                    if float(table['data']['render'][0]['render submissions'][0]['render shadow map'][2]):
                        shadows = True
                except (KeyError, TypeError):
                    pass

                try:
                    if float(table['data']['render'][0]['render submissions'][0]['render reflections'][2]):
                        fancywater = True
                except (KeyError, TypeError):
                    pass

        if msecs is None:
            return None

        options = []
        if shadows: options.append('S')
        if fancywater: options.append('W')

        return {
            'msecs': msecs,
            'map': mapname,
            'time': json['time'],
            'options': '[%s]' % '+'.join(options),
#            'json': json
        }

    profiles = []
    last_hwdetect = {}
    for report in reports:
        if report.data_type == 'hwdetect':
            last_hwdetect[report.user_id_hash] = summarise_hwdetect(report.downcast())
        elif report.data_type == 'profile':
            if report.user_id_hash in last_hwdetect:
                hwdetect = last_hwdetect[report.user_id_hash]
                if hwdetect['build_debug']:
                    continue
                prof = summarise_profile(report)
                if prof is not None:
                    profiles.append([report.user_id_hash, hwdetect, prof])

    datapoints = {}
    for user,hwdetect,profile in profiles:
        if profile['map'] != 'Death Canyon':
            continue
        if profile['time'] != 5:
            continue
        if profile['msecs'] is None or float(profile['msecs']) == 0:
            continue
        fps = 1000.0 / profile['msecs']
        title = '%s %s' % (hwdetect['device'], profile['options'])
        datapoints.setdefault(title, []).append(fps)

#    return render_to_response('reports/performance.html', {'data': datapoints, 'reports': profiles})

    sorted_datapoints = sorted(datapoints.items(), key = lambda (k,v):
        -numpy.median(v)
    )

    print "# %d datapoints" % sum(len(v) for k,v in sorted_datapoints)

    data_boxplot = [ v for k,v in sorted_datapoints ]
    data_scatter = ([], [])
    for i in range(len(data_boxplot)):
        for x in data_boxplot[i]:
            data_scatter[0].append(i+1)
            data_scatter[1].append(x)

    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    from matplotlib.figure import Figure

    fig = Figure(figsize=(16, 0.25*len(datapoints.items())))

    ax = fig.add_subplot(111)
    fig.subplots_adjust(left = 0.22, right = 0.98, top = 0.98, bottom = 0.05)

    ax.grid(True)

    ax.boxplot(data_boxplot, vert = 0, sym = '')
    ax.scatter(data_scatter[1], data_scatter[0], marker='x')

    ax.set_xlim(0.1, 1000)
    ax.set_xscale('log')
    ax.set_xlabel('Framerate (fps)')

    ax.set_yticklabels([k for k,v in sorted_datapoints], fontsize=8)
    ax.set_ylabel('Device [options: Shadows + Water reflections]')

    canvas = FigureCanvas(fig)
    response = HttpResponse(content_type = 'image/png')
    canvas.print_png(response, dpi=80)
    return response

def report_ram(request):
    reports = UserReport.objects
    reports = reports.filter(data_type = 'hwdetect', data_version__gte = 1)

    counts = {}
    for report in reports:
        #if not report.data_json()['os_linux']: continue
        ram = report.data_json()['ram_total']
        counts.setdefault(ram, set()).add(report.user_id_hash)

    datapoints = []
    accum = 0
    for size,count in sorted(counts.items()):
        accum += len(count)
        datapoints.append((size, accum))

    print "\n".join(repr(r) for r in datapoints)

    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    from matplotlib.figure import Figure

    fig = Figure(figsize=(16, 10))

    ax = fig.add_subplot(111)
    fig.subplots_adjust(left = 0.05, right = 0.98, top = 0.98, bottom = 0.05)

    ax.grid(True)

    ax.plot([ d[0] for d in datapoints ], [ 100*(1-float(d[1])/accum) for d in datapoints ])

    ax.set_xticks([0, 256, 512] + [1024*n for n in range(1, 9)])
    ax.set_xlim(0, 8192)
    ax.set_xlabel('RAM (megabytes)')

    ax.set_yticks(range(0, 101, 5))
    ax.set_ylim(0, 100)
    ax.set_ylabel('Cumulative percentage of users')

    canvas = FigureCanvas(fig)
    response = HttpResponse(content_type = 'image/png')
    canvas.print_png(response, dpi=80)
    return response

def report_os(request):
    reports = UserReport_hwdetect.objects
    reports = reports.filter(data_type = 'hwdetect', data_version__gte = 1)

    counts = {}
    for report in reports:
        json = report.data_json()
        if 'linux_release' in json:
            counts.setdefault(repr(json['linux_release']), set()).add(report.user_id_hash)
        os = report.os()
        counts.setdefault(os, set()).add(report.user_id_hash)

    out = ''
    for c in sorted(counts.items(), key = lambda c: len(c[1])):
        out += '%d %s\n' % (len(c[1]), c[0])

    return HttpResponse(out, content_type = 'text/plain')

def report_hwdetect_test_data(request):
    reports = UserReport_hwdetect.objects
    reports = reports.filter(data_type = 'hwdetect', data_version__gte = 1)

    data = set()
    for report in reports:
        json = report.data_json_nocache()
        relevant = {
            'os_unix': json['os_unix'],
            'os_linux': json['os_linux'],
            'os_macosx': json['os_macosx'],
            'os_win': json['os_win'],
            'gfx_card': json['gfx_card'],
            'gfx_drv_ver': json['gfx_drv_ver'],
            'gfx_mem': json['gfx_mem'],
            'GL_VENDOR': json['GL_VENDOR'],
            'GL_RENDERER': json['GL_RENDERER'],
            'GL_VERSION': json['GL_VERSION'],
            'GL_EXTENSIONS': json['GL_EXTENSIONS'],
        }
        data.add(hashabledict(relevant))

    json = simplejson.dumps(list(data), indent=1, sort_keys=True)
    return HttpResponse('var hwdetectTestData = %s' % json, content_type = 'text/plain')
